#include <sys/resource.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <signal.h>
#include <stdint.h>
#include <dlfcn.h>

#define BRK_LIMIT 100000 /* how much we allow the brk limit to grow */
static size_t brk_limit = 100000; /* brk increase required by default */

#define RAM_LIMIT 40000 /* heap RAM allowed to be allocated */
static size_t ram_limit = 45000; /* how much heap RAM we allocate by default */
static size_t maxi;  /* currently allocated RAM in the test, must stay
			below RAM_LIMIT; not really useful here, but
			may be in case of successive malloc calls */ 
static int flag = 0; /* 1 if we are in RAM control mode */

static long cycles = 100000l; /* default iterations to burn CPU time */

#define RTLD_NEXT ((void *) -1l)

/* will replace ALL calls to malloc(), so flag is there to localize the
   limit test; large malloc() allocation can happen outside the heap,
   so this is not bullet proof */

void * malloc(size_t size) { 
  static void * (*func)(size_t) = NULL; void * ret;

  if (!func) {
    /* get reference to original (libc provided) malloc */
    func = (void *(*)(size_t)) dlsym(RTLD_NEXT, "malloc");
  }

  if (flag  && (size+maxi > RAM_LIMIT)) {
    fprintf (stderr, "Decrease -m argument to pass! (was: %ld)\n", ram_limit);
    kill(getpid(),SIGSEGV);
  }

  /* call original malloc */
  ret = func(size);
  if (flag) maxi += size;
  
  return ret;
}

int test(void)
{
  struct rlimit r;

  printf("    brk:                      %10p\n", sbrk(0));

  getrlimit(RLIMIT_DATA, &r);
  r.rlim_cur = BRK_LIMIT;
  setrlimit(RLIMIT_DATA, &r);
  {
    void * p = sbrk(brk_limit);
    printf("    new brk:                  %10p\n", sbrk(0));
    if (p == (void *)-1) {
      fprintf (stderr, "Decrease -b argument to pass! (was: %ld)\n", brk_limit);
      kill(getpid(),SIGSEGV);
    }
    printf("    limit:                     %d bytes\n", sbrk(0)-p);
  }    

  maxi = 0;
  flag = 1;
  {
    /* should be locked if the new malloc is used by parent */
    char * p = malloc(ram_limit);
  }
  flag = 0;

  /* stack limit could be studied similarly using RLIMIT_STACK */
  
  /* RLIMIT_CPU is too coarse */
  
  {
    long i;
    struct itimerval interval;
    
    interval.it_value.tv_sec = 0;
    interval.it_value.tv_usec = 100;
    interval.it_interval.tv_sec = 0;
    interval.it_interval.tv_usec = 0;

    setitimer(ITIMER_PROF, &interval, NULL);

    /* compiler not clever enough to optimize that away */
    for (i = 0; i < cycles; ++i) {
      sqrt(drand48());
    }
  }
  fprintf (stderr, "Increase -c argument to fail! (was: %ld)\n", cycles);
  exit(EXIT_SUCCESS);
}

int main(int argc, char *argv[])
{
  char *usage = "Usage: limit [-m ram] [-b brk] [-c cycles]\n";
  char *p;

  while (argc > 1 && *argv[1] == '-')
  {
    if (*(argv[1] + 1) == 'm') {
      if (argc > 2) {
	errno = 0;
	ram_limit = strtol(argv[2], &p, 10);
	if ((*p != '\0') || (errno != 0)) {
	  fprintf (stderr, "Bad number after -m\n\n%s", usage);
	  exit (EXIT_FAILURE);
	}
      } else {
        fprintf (stderr, "Missing parameter for -m\n\n%s", usage);
        exit (EXIT_FAILURE);
      }
      argc--;
      argv++;
    } else if (*(argv[1] + 1) == 'b') {
      if (argc > 2) {
	errno = 0;
	brk_limit = strtol(argv[2], &p, 10);
	if ((*p != '\0') || (errno != 0)) {
	  fprintf (stderr, "Bad number after -b\n\n%s", usage);
	  exit (EXIT_FAILURE);
	}
      } else {
        fprintf (stderr, "Missing parameter for -b\n\n%s",usage);
        exit (EXIT_FAILURE);
      }
      argc--;
      argv++;
    } else if (*(argv[1] + 1) == 'c') {
      if (argc > 2) {
	errno = 0;
	cycles = strtol(argv[2], &p, 10);
	if ((*p != '\0') || (errno != 0)) {
	  fprintf (stderr, "Bad number after -c\n\n%s", usage);
	  exit (EXIT_FAILURE);
	}
      } else {
        fprintf (stderr, "Missing parameter for -c\n\n%s", usage);
        exit (EXIT_FAILURE);
      }
      argc--;
      argv++;
    } else {
        fprintf (stderr, "Invalid parameter %s\n\n%s", argv[1], usage);
        exit (EXIT_FAILURE);
    }
    argc--;
    argv++;
  }
  
  {
    pid_t pid = fork();

    if (pid < 0) {
      fprintf(stderr, "Fork failure, exiting\n");    
      exit(EXIT_FAILURE);
    }

    if (pid) {
      int status;
      pid_t pidw = wait(&status);

      if (pidw != pid) {
	fprintf(stderr, "Bad child died, exiting\n");    
	exit(EXIT_FAILURE);
      }

      if (WIFEXITED(status)) {
	fprintf(stderr, "Executed within limits\n");
	exit(EXIT_SUCCESS);
      }
      if (WIFSIGNALED(status)) {
	fprintf(stderr, "Failure executing within limits: %s\n", strsignal(WTERMSIG(status)));
	exit(EXIT_SUCCESS);
      }
      fprintf(stderr, "WTF?\n");
      exit(EXIT_FAILURE);
    } else {
      fprintf(stderr, "Child: testing running within limits...\n");
      test();
    }
  }
}

